/* This file is part of Cloudy and is copyright (C)1978-2022 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */

#ifndef ATMDAT_ADFA_H_
#define ATMDAT_ADFA_H_

const int NRECCOEFCNO=471;
typedef enum { PHFIT_UNDEF, PHFIT95, PHFIT96 } phfit_version;

class t_ADfA : public Singleton<t_ADfA>
{
	friend class Singleton<t_ADfA>;
protected: 
	t_ADfA();
private:
	phfit_version version;
	/* phfit.dat */
	long int L[7];
	long int NINN[30];
	long int NTOT[30];
	realnum PH1[7][30][30][6];
	realnum PH2[30][30][7];
	/* hpfit.dat */
	realnum PHH[NHYDRO_MAX_LEVEL][5];
	/* rec_lines.dat */
	realnum P[8][110];
	realnum ST[9][405];
	/* rad_rec.dat */
	realnum rrec[30][30][2];
	realnum rnew[30][30][4];
	realnum fe[13][3];
	/* h_rad_rec */
	realnum HRF[NHYDRO_MAX_LEVEL][9];
	/* h_phot_cs.dat */
	/** array of cross sections for photoionization of hydrogen at threshold,
	 * 0 is 1s, 1 is 2s, 2 is 2p, up to 400 */
	realnum STH[NHYDRO_MAX_LEVEL];
	/* coll_ion.dat */
	double CF[30][30][5];
	/* h_coll_str.dat */
	/** array of EIE cross sections for hydrogen atom.  
	 * For all E1 transitions nl - n'l', with n' < n <= 5 */
	/* >>refer	H1	cs	Anderson, H., Ballance, C.P., Badnell, N.R., & Summers, H.P., 
	 * >>refercon	2000, J Phys B, 33, 1255; erratum, 2002 */
	double HCS[14][10][8];
public:
	/** set_version set version of phfit data to be used
	    \param val
	*/
	void set_version(phfit_version val) { version = val; }

	/** get_version which version of phfit data should be used? */
	phfit_version get_version() const { return version; }

	/** ph1 access elements of PH1 data block with parameters for photoionization cross section fits
	    \param i
	    \param j
	    \param k
	    \param l
	*/
	realnum ph1(int i, int j, int k, int l) const { return PH1[i][j][k][l]; }

	/** sth array of cross sections for photoionization of hydrogen at threshold,
	    0 is 1s, 1 is 2s, 2 is 2p, up to 400
	    \param i
	*/
	realnum sth(int i) const { return STH[i]; }

	/** phfit this subroutine calculates partial photoionization cross sections
	    for all ionization stages of all atoms from H to Zn (Z=30)
	    \param nz
	    \param ne
	    \param is
	    \param e
	    \author Dima Verner
	*/
	double phfit(long int nz, long int ne, long int is, double e);

	/** hpfit state specific photoionization cross sections for model hydrogen atom
	    \param iz 
	    \param n 
	    \param e 
	    \author Dima Verner
	*/ 
	double hpfit(long int iz, long int n, double e);

	/** rec_lines effective recombination coefficients for lines of C, N, O, by D. Verner
	    \param  t 
	    \param  r
	    \author Dima Verner
	*/ 
	void rec_lines(double t, realnum r[][NRECCOEFCNO]);

	/** rad_rec calculates rates of radiative recombination for all ions
	    \param iz nuclear number on physics scale
	    \param in number of recombined electrons
	    \param t temperature K
	    \author Dima Verner
	*/ 
	double rad_rec(long int iz, long int in, double t);

	/** H_rad_rec calculates state-specific recombination rates for H and H-like ions
	    \param iz 
	    \param n 
	    \param t
	    \author Dima Verner
	*/ 
	double H_rad_rec(long int iz, long int n, double t);

	/** coll_ion D Verner's routine to compute collisional ionization rate coefficients 
	    \param iz 
	    \param in 
	    \param t
	    \author Dima Verner
	*/
	double coll_ion(long int iz, long int in, double t);

	double coll_ion_wrapper(long int z, long int n, double t);

	/* coll_ion_hybrid computes hybrid collisional ionization rates */
	double coll_ion_hybrid(	long int z, long int n,	double t);

	/** h_coll_str routine to grab H cross sections from Anderson et al. 2002.
	    \param ipLo 
	    \param ipHi 
	    \param ipTe
	*/
	realnum h_coll_str( long ipLo, long ipHi, long ipTe );
};

#endif
